@@ -1,5 +1,5 @@
 /*  Minilzip - Test program for the lzlib library
-    Copyright (C) 2009-2014 Antonio Diaz Diaz.
+    Copyright (C) 2009-2017 Antonio Diaz Diaz.
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -71,15 +71,18 @@
 
 void cleanup_and_fail( const int retval );
 void show_error( const char * const msg, const int errcode, const bool help );
+void show_file_error( const char * const filename, const char * const msg,
+                      const int errcode );
 void internal_error( const char * const msg );
 
+int verbosity = 0;
 
 const char * const Program_name = "Minilzip";
 const char * const program_name = "minilzip";
-const char * const program_year = "2014";
+const char * const program_year = "2017";
 const char * invocation_name = 0;
 
-struct { const char * from; const char * to; } const known_extensions[] = {
+const struct { const char * from; const char * to; } known_extensions[] = {
   { ".lz",  ""     },
   { ".tlz", ".tar" },
   { 0,      0      } };
@@ -94,10 +97,6 @@ enum Mode { m_compress, m_decompress, m_test };
 
 char * output_filename = 0;
 int outfd = -1;
-int verbosity = 0;
-const mode_t usr_rw = S_IRUSR | S_IWUSR;
-const mode_t all_rw = S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH;
-mode_t outfd_mode = S_IRUSR | S_IWUSR;
 bool delete_output_on_interrupt = false;
 
 
@@ -110,7 +109,8 @@ struct Pretty_print
   };
 
 static void Pp_init( struct Pretty_print * const pp,
-                     const char * const filenames[], const int num_filenames )
+                     const char * const filenames[],
+                     const int num_filenames, const int verbosity )
   {
   unsigned stdin_name_len;
   int i;
@@ -118,8 +118,9 @@ static void Pp_init( struct Pretty_print * const pp,
   pp->stdin_name = "(stdin)";
   pp->longest_name = 0;
   pp->first_post = false;
-  stdin_name_len = strlen( pp->stdin_name );
 
+  if( verbosity <= 0 ) return;
+  stdin_name_len = strlen( pp->stdin_name );
   for( i = 0; i < num_filenames; ++i )
     {
     const char * const s = filenames[i];
@@ -147,11 +148,11 @@ static void Pp_show_msg( struct Pretty_print * const pp, const char * const msg
     {
     if( pp->first_post )
       {
-      int i, len;
+      unsigned i;
       pp->first_post = false;
       fprintf( stderr, "  %s: ", pp->name );
-      len = pp->longest_name - strlen( pp->name );
-      for( i = 0; i < len; ++i ) fprintf( stderr, " " );
+      for( i = strlen( pp->name ); i < pp->longest_name; ++i )
+        fputc( ' ', stderr );
       if( !msg ) fflush( stderr );
       }
     if( msg ) fprintf( stderr, "%s\n", msg );
@@ -166,31 +167,33 @@ static void show_help( void )
   printf( "\nOptions:\n"
           "  -h, --help                     display this help and exit\n"
           "  -V, --version                  output version information and exit\n"
+          "  -a, --trailing-error           exit with error status if trailing data\n"
           "  -b, --member-size=<bytes>      set member size limit in bytes\n"
-          "  -c, --stdout                   send output to standard output\n"
+          "  -c, --stdout                   write to standard output, keep input files\n"
           "  -d, --decompress               decompress\n"
           "  -f, --force                    overwrite existing output files\n"
-          "  -F, --recompress               force recompression of compressed files\n"
+          "  -F, --recompress               force re-compression of compressed files\n"
           "  -k, --keep                     keep (don't delete) input files\n"
           "  -m, --match-length=<bytes>     set match length limit in bytes [36]\n"
-          "  -o, --output=<file>            if reading stdin, place the output into <file>\n"
+          "  -o, --output=<file>            if reading standard input, write to <file>\n"
           "  -q, --quiet                    suppress all messages\n"
           "  -s, --dictionary-size=<bytes>  set dictionary size limit in bytes [8 MiB]\n"
           "  -S, --volume-size=<bytes>      set volume size limit in bytes\n"
           "  -t, --test                     test compressed file integrity\n"
           "  -v, --verbose                  be verbose (a 2nd -v gives more)\n"
-          "  -1 .. -9                       set compression level [default 6]\n"
-          "      --fast                     alias for -1\n"
+          "  -0 .. -9                       set compression level [default 6]\n"
+          "      --fast                     alias for -0\n"
           "      --best                     alias for -9\n"
-          "If no file names are given, minilzip compresses or decompresses\n"
-          "from standard input to standard output.\n"
+          "If no file names are given, or if a file is '-', minilzip compresses or\n"
+          "decompresses from standard input to standard output.\n"
           "Numbers may be followed by a multiplier: k = kB = 10^3 = 1000,\n"
           "Ki = KiB = 2^10 = 1024, M = 10^6, Mi = 2^20, G = 10^9, Gi = 2^30, etc...\n"
-          "The bidimensional parameter space of LZMA can't be mapped to a linear\n"
+          "Dictionary sizes 12 to 29 are interpreted as powers of two, meaning 2^12\n"
+          "to 2^29 bytes.\n"
+          "\nThe bidimensional parameter space of LZMA can't be mapped to a linear\n"
           "scale optimal for all files. If your files are large, very repetitive,\n"
-          "etc, you may need to use the --match-length and --dictionary-size\n"
-          "options directly to achieve optimal performance. For example, -9m64\n"
-          "usually compresses executables more (and faster) than -9.\n"
+          "etc, you may need to use the --dictionary-size and --match-length\n"
+          "options directly to achieve optimal performance.\n"
           "\nExit status: 0 for a normal exit, 1 for environmental problems (file\n"
           "not found, invalid flags, I/O errors, etc), 2 to indicate a corrupt or\n"
           "invalid input file, 3 for an internal consistency error (eg, bug) which\n"
@@ -213,18 +216,21 @@ static void show_version( void )
 
 static void show_header( const unsigned dictionary_size )
   {
-  const char * const prefix[8] =
-    { "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi", "Yi" };
-  enum { factor = 1024 };
-  const char * p = "";
-  const char * np = "  ";
-  unsigned num = dictionary_size, i;
-  bool exact = ( num % factor == 0 );
+  if( verbosity >= 3 )
+    {
+    enum { factor = 1024 };
+    const char * const prefix[8] =
+      { "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi", "Yi" };
+    const char * p = "";
+    const char * np = "  ";
+    unsigned num = dictionary_size, i;
+    bool exact = ( num % factor == 0 );
 
-  for( i = 0; i < 8 && ( num > 9999 || ( exact && num >= factor ) ); ++i )
-    { num /= factor; if( num % factor != 0 ) exact = false;
-      p = prefix[i]; np = ""; }
-  fprintf( stderr, "dictionary size %s%4u %sB.  ", np, num, p );
+    for( i = 0; i < 8 && ( num > 9999 || ( exact && num >= factor ) ); ++i )
+      { num /= factor; if( num % factor != 0 ) exact = false;
+        p = prefix[i]; np = ""; }
+    fprintf( stderr, "dictionary %s%4u %sB.  ", np, num, p );
+    }
   }
 
 
@@ -244,12 +250,11 @@ static unsigned long long getnum( const char * const ptr,
 
   if( !errno && tail[0] )
     {
-    int factor = ( tail[1] == 'i' ) ? 1024 : 1000;
-    int exponent = 0, i;
-    bool bad_multiplier = false;
+    const unsigned factor = ( tail[1] == 'i' ) ? 1024 : 1000;
+    int exponent = 0;				/* 0 = bad multiplier */
+    int i;
     switch( tail[0] )
       {
-      case ' ': break;
       case 'Y': exponent = 8; break;
       case 'Z': exponent = 7; break;
       case 'E': exponent = 6; break;
@@ -257,13 +262,10 @@ static unsigned long long getnum( const char * const ptr,
       case 'T': exponent = 4; break;
       case 'G': exponent = 3; break;
       case 'M': exponent = 2; break;
-      case 'K': if( factor == 1024 ) exponent = 1; else bad_multiplier = true;
-                break;
-      case 'k': if( factor == 1000 ) exponent = 1; else bad_multiplier = true;
-                break;
-      default : bad_multiplier = true;
+      case 'K': if( factor == 1024 ) exponent = 1; break;
+      case 'k': if( factor == 1000 ) exponent = 1; break;
       }
-    if( bad_multiplier )
+    if( exponent <= 0 )
       {
       show_error( "Bad multiplier in numerical argument.", 0, true );
       exit( 1 );
@@ -287,23 +289,29 @@ static unsigned long long getnum( const char * const ptr,
 static int get_dict_size( const char * const arg )
   {
   char * tail;
-  int bits = strtol( arg, &tail, 0 );
+  int dictionary_size;
+  const long bits = strtol( arg, &tail, 0 );
   if( bits >= LZ_min_dictionary_bits() &&
       bits <= LZ_max_dictionary_bits() && *tail == 0 )
     return ( 1 << bits );
-  return getnum( arg, LZ_min_dictionary_size(), LZ_max_dictionary_size() );
+  dictionary_size = getnum( arg, LZ_min_dictionary_size(),
+                                 LZ_max_dictionary_size() );
+  if( dictionary_size == 65535 ) ++dictionary_size;	/* no fast encoder */
+  return dictionary_size;
   }
 
 
 static int extension_index( const char * const name )
   {
-  int i;
-  for( i = 0; known_extensions[i].from; ++i )
+  int eindex;
+  for( eindex = 0; known_extensions[eindex].from; ++eindex )
     {
-    const char * const ext = known_extensions[i].from;
-    if( strlen( name ) > strlen( ext ) &&
-        strncmp( name + strlen( name ) - strlen( ext ), ext, strlen( ext ) ) == 0 )
-      return i;
+    const char * const ext = known_extensions[eindex].from;
+    const unsigned name_len = strlen( name );
+    const unsigned ext_len = strlen( ext );
+    if( name_len > ext_len &&
+        strncmp( name + name_len - ext_len, ext, ext_len ) == 0 )
+      return eindex;
     }
   return -1;
   }
@@ -324,11 +332,7 @@ static int open_instream( const char * const name, struct stat * const in_statsp
     {
     infd = open( name, O_RDONLY | O_BINARY );
     if( infd < 0 )
-      {
-      if( verbosity >= 0 )
-        fprintf( stderr, "%s: Can't open input file '%s': %s.\n",
-                 program_name, name, strerror( errno ) );
-      }
+      show_file_error( name, "Can't open input file", errno );
     else
       {
       const int i = fstat( infd, in_statsp );
@@ -336,14 +340,14 @@ static int open_instream( const char * const name, struct stat * const in_statsp
       const bool can_read = ( i == 0 &&
                               ( S_ISBLK( mode ) || S_ISCHR( mode ) ||
                                 S_ISFIFO( mode ) || S_ISSOCK( mode ) ) );
-      const bool no_ofile = to_stdout || program_mode == m_test;
+      const bool no_ofile = ( to_stdout || program_mode == m_test );
       if( i != 0 || ( !S_ISREG( mode ) && ( !can_read || !no_ofile ) ) )
         {
         if( verbosity >= 0 )
           fprintf( stderr, "%s: Input file '%s' is not a regular file%s.\n",
                    program_name, name,
                    ( can_read && !no_ofile ) ?
-                   " and '--stdout' was not specified" : "" );
+                   ",\n          and '--stdout' was not specified" : "" );
         close( infd );
         infd = -1;
         }
@@ -354,7 +358,7 @@ static int open_instream( const char * const name, struct stat * const in_statsp
 
 
 /* assure at least a minimum size for buffer 'buf' */
-static void * resize_buffer( void * buf, const int min_size )
+static void * resize_buffer( void * buf, const unsigned min_size )
   {
   if( buf ) buf = realloc( buf, min_size );
   else buf = malloc( min_size );
@@ -377,52 +381,58 @@ static void set_c_outname( const char * const name, const bool multifile )
   }
 
 
-static void set_d_outname( const char * const name, const int i )
+static void set_d_outname( const char * const name, const int eindex )
   {
-  if( i >= 0 )
+  const unsigned name_len = strlen( name );
+  if( eindex >= 0 )
     {
-    const char * const from = known_extensions[i].from;
-    if( strlen( name ) > strlen( from ) )
+    const char * const from = known_extensions[eindex].from;
+    const unsigned from_len = strlen( from );
+    if( name_len > from_len )
       {
-      output_filename = resize_buffer( output_filename, strlen( name ) +
-                                       strlen( known_extensions[0].to ) + 1 );
+      output_filename = resize_buffer( output_filename, name_len +
+                                       strlen( known_extensions[eindex].to ) + 1 );
       strcpy( output_filename, name );
-      strcpy( output_filename + strlen( name ) - strlen( from ),
-              known_extensions[i].to );
+      strcpy( output_filename + name_len - from_len, known_extensions[eindex].to );
       return;
       }
     }
-  output_filename = resize_buffer( output_filename, strlen( name ) + 4 + 1 );
+  output_filename = resize_buffer( output_filename, name_len + 4 + 1 );
   strcpy( output_filename, name );
   strcat( output_filename, ".out" );
   if( verbosity >= 1 )
-    fprintf( stderr, "%s: Can't guess original name for '%s' -- using '%s'.\n",
+    fprintf( stderr, "%s: Can't guess original name for '%s' -- using '%s'\n",
              program_name, name, output_filename );
   }
 
 
-static bool open_outstream( const bool force )
+static bool open_outstream( const bool force, const bool from_stdin )
   {
+  const mode_t usr_rw = S_IRUSR | S_IWUSR;
+  const mode_t all_rw = usr_rw | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH;
+  const mode_t outfd_mode = from_stdin ? all_rw : usr_rw;
   int flags = O_CREAT | O_WRONLY | O_BINARY;
   if( force ) flags |= O_TRUNC; else flags |= O_EXCL;
 
   outfd = open( output_filename, flags, outfd_mode );
-  if( outfd < 0 && verbosity >= 0 )
+  if( outfd >= 0 ) delete_output_on_interrupt = true;
+  else if( verbosity >= 0 )
     {
     if( errno == EEXIST )
       fprintf( stderr, "%s: Output file '%s' already exists, skipping.\n",
                program_name, output_filename );
     else
-      fprintf( stderr, "%s: Can't create output file '%s': %s.\n",
+      fprintf( stderr, "%s: Can't create output file '%s': %s\n",
                program_name, output_filename, strerror( errno ) );
     }
   return ( outfd >= 0 );
   }
 
 
-static bool check_tty( const int infd, const enum Mode program_mode )
+static bool check_tty( const char * const input_filename, const int infd,
+                       const enum Mode program_mode )
   {
-  if( program_mode == m_compress && outfd >= 0 && isatty( outfd ) )
+  if( program_mode == m_compress && isatty( outfd ) )
     {
     show_error( "I won't write compressed data to a terminal.", 0, true );
     return false;
@@ -430,7 +440,8 @@ static bool check_tty( const int infd, const enum Mode program_mode )
   if( ( program_mode == m_decompress || program_mode == m_test ) &&
       isatty( infd ) )
     {
-    show_error( "I won't read compressed data from a terminal.", 0, true );
+    show_file_error( input_filename,
+                     "I won't read compressed data from a terminal.", 0 );
     return false;
     }
   return true;
@@ -468,7 +479,11 @@ static void close_and_set_permissions( const struct stat * const in_statsp )
           fchmod( outfd, mode & ~( S_ISUID | S_ISGID | S_ISVTX ) ) != 0 )
         warning = true;
     }
-  if( close( outfd ) != 0 ) cleanup_and_fail( 1 );
+  if( close( outfd ) != 0 )
+    {
+    show_error( "Error closing output file", errno, false );
+    cleanup_and_fail( 1 );
+    }
   outfd = -1;
   delete_output_on_interrupt = false;
   if( in_statsp )
@@ -522,11 +537,11 @@ static int writeblock( const int fd, const uint8_t * const buf, const int size )
 
 static bool next_filename( void )
   {
-  const unsigned len = strlen( known_extensions[0].from );
+  const unsigned name_len = strlen( output_filename );
+  const unsigned ext_len = strlen( known_extensions[0].from );
   int i, j;
-
-  if( strlen( output_filename ) >= len + 5 )		/* "*00001.lz" */
-    for( i = strlen( output_filename ) - len - 1, j = 0; j < 5; --i, ++j )
+  if( name_len >= ext_len + 5 )				/* "*00001.lz" */
+    for( i = name_len - ext_len - 1, j = 0; j < 5; --i, ++j )
       {
       if( output_filename[i] < '9' ) { ++output_filename[i]; return true; }
       else output_filename[i] = '0';
@@ -544,8 +559,8 @@ static int do_compress( struct LZ_Encoder * const encoder,
   unsigned long long partial_volume_size = 0;
   enum { buffer_size = 65536 };
   uint8_t buffer[buffer_size];
-
   if( verbosity >= 1 ) Pp_show_msg( pp, 0 );
+
   while( true )
     {
     int in_size = 0, out_size;
@@ -569,9 +584,9 @@ static int do_compress( struct LZ_Encoder * const encoder,
       {
       Pp_show_msg( pp, 0 );
       if( verbosity >= 0 )
-        fprintf( stderr, "%s: LZ_compress_read error: %s.\n",
+        fprintf( stderr, "%s: LZ_compress_read error: %s\n",
                  program_name, LZ_strerror( LZ_compress_errno( encoder ) ) );
-        return 1;
+      return 1;
       }
     else if( out_size > 0 )
       {
@@ -598,8 +613,7 @@ static int do_compress( struct LZ_Encoder * const encoder,
             close_and_set_permissions( in_statsp );
             if( !next_filename() )
               { Pp_show_msg( pp, "Too many volume files." ); return 1; }
-            if( !open_outstream( true ) ) return 1;
-            delete_output_on_interrupt = true;
+            if( !open_outstream( true, !in_statsp ) ) return 1;
             }
           }
         size = min( member_size, volume_size - partial_volume_size );
@@ -610,7 +624,7 @@ static int do_compress( struct LZ_Encoder * const encoder,
         {
         Pp_show_msg( pp, 0 );
         if( verbosity >= 0 )
-          fprintf( stderr, "%s: LZ_compress_restart_member error: %s.\n",
+          fprintf( stderr, "%s: LZ_compress_restart_member error: %s\n",
                    program_name, LZ_strerror( LZ_compress_errno( encoder ) ) );
         return 1;
         }
@@ -622,7 +636,7 @@ static int do_compress( struct LZ_Encoder * const encoder,
     const unsigned long long in_size = LZ_compress_total_in_size( encoder );
     const unsigned long long out_size = LZ_compress_total_out_size( encoder );
     if( in_size == 0 || out_size == 0 )
-      fprintf( stderr, " no data compressed.\n" );
+      fputs( " no data compressed.\n", stderr );
     else
       fprintf( stderr, "%6.3f:1, %6.3f bits/byte, "
                        "%5.2f%% saved, %llu in, %llu out.\n",
@@ -635,11 +649,11 @@ static int do_compress( struct LZ_Encoder * const encoder,
   }
 
 
-int compress( const unsigned long long member_size,
-              const unsigned long long volume_size,
-              const struct Lzma_options * const encoder_options,
-              const int infd, struct Pretty_print * const pp,
-              const struct stat * const in_statsp )
+static int compress( const unsigned long long member_size,
+                     const unsigned long long volume_size, const int infd,
+                     const struct Lzma_options * const encoder_options,
+                     struct Pretty_print * const pp,
+                     const struct stat * const in_statsp )
   {
   struct LZ_Encoder * const encoder =
     LZ_compress_open( encoder_options->dictionary_size,
@@ -662,8 +676,9 @@ int compress( const unsigned long long member_size,
   }
 
 
-int do_decompress( struct LZ_Decoder * const decoder, const int infd,
-                   struct Pretty_print * const pp, const bool testing )
+static int do_decompress( struct LZ_Decoder * const decoder, const int infd,
+                          struct Pretty_print * const pp,
+                          const bool ignore_trailing, const bool testing )
   {
   enum { buffer_size = 65536 };
   uint8_t buffer[buffer_size];
@@ -671,11 +686,10 @@ int do_decompress( struct LZ_Decoder * const decoder, const int infd,
 
   for( first_member = true; ; )
     {
-    int in_size = min( LZ_decompress_write_size( decoder ), buffer_size );
-    int out_size = 0;
-    if( in_size > 0 )
+    const int max_in_size = min( LZ_decompress_write_size( decoder ), buffer_size );
+    int in_size = 0, out_size = 0;
+    if( max_in_size > 0 )
       {
-      const int max_in_size = in_size;
       in_size = readblock( infd, buffer, max_in_size );
       if( in_size != max_in_size && errno )
         {
@@ -707,21 +721,19 @@ int do_decompress( struct LZ_Decoder * const decoder, const int infd,
         {
         if( verbosity >= 1 )
           {
-          const unsigned long long data_position = LZ_decompress_data_position( decoder );
+          const unsigned long long data_size = LZ_decompress_data_position( decoder );
           const unsigned long long member_size = LZ_decompress_member_position( decoder );
           Pp_show_msg( pp, 0 );
-          if( verbosity >= 3 )
-            show_header( LZ_decompress_dictionary_size( decoder ) );
-          if( verbosity >= 2 && data_position > 0 && member_size > 0 )
+          show_header( LZ_decompress_dictionary_size( decoder ) );
+          if( verbosity >= 2 && data_size > 0 && member_size > 0 )
             fprintf( stderr, "%6.3f:1, %6.3f bits/byte, %5.2f%% saved.  ",
-                     (double)data_position / member_size,
-                     ( 8.0 * member_size ) / data_position,
-                     100.0 * ( 1.0 - ( (double)member_size / data_position ) ) );
+                     (double)data_size / member_size,
+                     ( 8.0 * member_size ) / data_size,
+                     100.0 * ( 1.0 - ( (double)member_size / data_size ) ) );
           if( verbosity >= 4 )
-            fprintf( stderr, "data CRC %08X, data size %9llu, member size %8llu.  ",
-                     LZ_decompress_data_crc( decoder ),
-                     data_position, member_size );
-          fprintf( stderr, testing ? "ok\n" : "done\n" );
+            fprintf( stderr, "CRC %08X, decompressed %9llu, compressed %8llu.  ",
+                     LZ_decompress_data_crc( decoder ), data_size, member_size );
+          fputs( testing ? "ok\n" : "done\n", stderr );
           }
         first_member = false; Pp_reset( pp );
         }
@@ -730,11 +742,20 @@ int do_decompress( struct LZ_Decoder * const decoder, const int infd,
     if( out_size < 0 || ( first_member && out_size == 0 ) )
       {
       const enum LZ_Errno lz_errno = LZ_decompress_errno( decoder );
-      if( lz_errno == LZ_header_error || ( first_member && out_size == 0 ) )
+      if( lz_errno == LZ_unexpected_eof &&
+          LZ_decompress_member_position( decoder ) <= 6 )
+        { Pp_show_msg( pp, "File ends unexpectedly at member header." );
+          return 2; }
+      if( lz_errno == LZ_header_error )
         {
-        if( !first_member ) break;		/* trailing garbage */
-        Pp_show_msg( pp, "Bad magic number (file not in lzip format)." );
-        return 2;
+        if( first_member )
+          { show_file_error( pp->name,
+                             "Bad magic number (file not in lzip format).", 0 );
+            return 2; }
+        else if( !ignore_trailing )
+          { show_file_error( pp->name, "Trailing data not allowed.", 0 );
+            return 2; }
+        break;
         }
       if( lz_errno == LZ_mem_error )
         { Pp_show_msg( pp, "Not enough memory." ); return 1; }
@@ -742,10 +763,10 @@ int do_decompress( struct LZ_Decoder * const decoder, const int infd,
         {
         Pp_show_msg( pp, 0 );
         if( lz_errno == LZ_unexpected_eof )
-          fprintf( stderr, "File ends unexpectedly at pos %llu.\n",
+          fprintf( stderr, "File ends unexpectedly at pos %llu\n",
                    LZ_decompress_total_in_size( decoder ) );
         else
-          fprintf( stderr, "Decoder error at pos %llu: %s.\n",
+          fprintf( stderr, "Decoder error at pos %llu: %s\n",
                    LZ_decompress_total_in_size( decoder ),
                    LZ_strerror( LZ_decompress_errno( decoder ) ) );
         }
@@ -759,15 +780,15 @@ int do_decompress( struct LZ_Decoder * const decoder, const int infd,
   }
 
 
-int decompress( const int infd, struct Pretty_print * const pp,
-                const bool testing )
+static int decompress( const int infd, struct Pretty_print * const pp,
+                       const bool ignore_trailing, const bool testing )
   {
   struct LZ_Decoder * const decoder = LZ_decompress_open();
   int retval;
 
   if( !decoder || LZ_decompress_errno( decoder ) != LZ_ok )
     { Pp_show_msg( pp, "Not enough memory." ); retval = 1; }
-  else retval = do_decompress( decoder, infd, pp, testing );
+  else retval = do_decompress( decoder, infd, pp, ignore_trailing, testing );
 
   LZ_decompress_close( decoder );
   return retval;
@@ -792,18 +813,26 @@ static void set_signals( void )
 
 void show_error( const char * const msg, const int errcode, const bool help )
   {
-  if( verbosity >= 0 )
+  if( verbosity < 0 ) return;
+  if( msg && msg[0] )
     {
-    if( msg && msg[0] )
-      {
-      fprintf( stderr, "%s: %s", program_name, msg );
-      if( errcode > 0 ) fprintf( stderr, ": %s.", strerror( errcode ) );
-      fprintf( stderr, "\n" );
-      }
-    if( help )
-      fprintf( stderr, "Try '%s --help' for more information.\n",
-               invocation_name );
+    fprintf( stderr, "%s: %s", program_name, msg );
+    if( errcode > 0 ) fprintf( stderr, ": %s", strerror( errcode ) );
+    fputc( '\n', stderr );
     }
+  if( help )
+    fprintf( stderr, "Try '%s --help' for more information.\n",
+             invocation_name );
+  }
+
+
+void show_file_error( const char * const filename, const char * const msg,
+                      const int errcode )
+  {
+  if( verbosity < 0 ) return;
+  fprintf( stderr, "%s: %s: %s", program_name, filename, msg );
+  if( errcode > 0 ) fprintf( stderr, ": %s", strerror( errcode ) );
+  fputc( '\n', stderr );
   }
 
 
@@ -821,7 +850,7 @@ int main( const int argc, const char * const argv[] )
      to the corresponding LZMA compression modes. */
   const struct Lzma_options option_mapping[] =
     {
-    { 1 << 20,   5 },		/* -0 */
+    {   65535,  16 },		/* -0 (65535,16 chooses fast encoder) */
     { 1 << 20,   5 },		/* -1 */
     { 3 << 19,   6 },		/* -2 */
     { 1 << 21,   8 },		/* -3 */
@@ -832,11 +861,10 @@ int main( const int argc, const char * const argv[] )
     { 3 << 23, 132 },		/* -8 */
     { 1 << 25, 273 } };		/* -9 */
   struct Lzma_options encoder_options = option_mapping[6];  /* default = "-6" */
-  const unsigned long long max_member_size = 0x0100000000000000ULL;
+  const unsigned long long max_member_size = 0x0008000000000000ULL;
   const unsigned long long max_volume_size = 0x4000000000000000ULL;
   unsigned long long member_size = max_member_size;
   unsigned long long volume_size = 0;
-  const char * input_filename = "";
   const char * default_output_filename = "";
   const char ** filenames = 0;
   int num_filenames = 0;
@@ -847,8 +875,10 @@ int main( const int argc, const char * const argv[] )
   int i;
   bool filenames_given = false;
   bool force = false;
+  bool ignore_trailing = true;
   bool keep_input_files = false;
   bool recompress = false;
+  bool stdin_used = false;
   bool to_stdout = false;
   struct Pretty_print pp;
 
@@ -864,6 +894,7 @@ int main( const int argc, const char * const argv[] )
     { '7',  0,                ap_no  },
     { '8',  0,                ap_no  },
     { '9', "best",            ap_no  },
+    { 'a', "trailing-error",  ap_no  },
     { 'b', "member-size",     ap_yes },
     { 'c', "stdout",          ap_no  },
     { 'd', "decompress",      ap_no  },
@@ -906,6 +937,7 @@ int main( const int argc, const char * const argv[] )
       case '0': case '1': case '2': case '3': case '4':
       case '5': case '6': case '7': case '8': case '9':
                 encoder_options = option_mapping[code-'0']; break;
+      case 'a': ignore_trailing = false; break;
       case 'b': member_size = getnum( arg, 100000, max_member_size ); break;
       case 'c': to_stdout = true; break;
       case 'd': program_mode = m_decompress; break;
@@ -934,9 +966,6 @@ int main( const int argc, const char * const argv[] )
   setmode( STDOUT_FILENO, O_BINARY );
 #endif
 
-  if( program_mode == m_test )
-    outfd = -1;
-
   num_filenames = max( 1, ap_arguments( &parser ) - argind );
   filenames = resize_buffer( filenames, num_filenames * sizeof filenames[0] );
   filenames[0] = "-";
@@ -947,15 +976,19 @@ int main( const int argc, const char * const argv[] )
     if( strcmp( filenames[i], "-" ) != 0 ) filenames_given = true;
     }
 
+  if( program_mode == m_test )
+    outfd = -1;
+
   if( !to_stdout && program_mode != m_test &&
       ( filenames_given || default_output_filename[0] ) )
     set_signals();
 
-  Pp_init( &pp, filenames, num_filenames );
+  Pp_init( &pp, filenames, num_filenames, verbosity );
 
   output_filename = resize_buffer( output_filename, 1 );
   for( i = 0; i < num_filenames; ++i )
     {
+    const char * input_filename = "";
     int tmp;
     struct stat in_stats;
     const struct stat * in_statsp;
@@ -963,7 +996,7 @@ int main( const int argc, const char * const argv[] )
 
     if( !filenames[i][0] || strcmp( filenames[i], "-" ) == 0 )
       {
-      input_filename = "";
+      if( stdin_used ) continue; else stdin_used = true;
       infd = STDIN_FILENO;
       if( program_mode != m_test )
         {
@@ -979,8 +1012,7 @@ int main( const int argc, const char * const argv[] )
                                              strlen( default_output_filename ) + 1 );
             strcpy( output_filename, default_output_filename );
             }
-          outfd_mode = all_rw;
-          if( !open_outstream( force ) )
+          if( !open_outstream( force, true ) )
             {
             if( retval < 1 ) retval = 1;
             close( infd ); infd = -1;
@@ -991,8 +1023,7 @@ int main( const int argc, const char * const argv[] )
       }
     else
       {
-      const int eindex = extension_index( filenames[i] );
-      input_filename = filenames[i];
+      const int eindex = extension_index( input_filename = filenames[i] );
       infd = open_instream( input_filename, &in_stats, program_mode,
                             eindex, recompress, to_stdout );
       if( infd < 0 ) { if( retval < 1 ) retval = 1; continue; }
@@ -1004,8 +1035,7 @@ int main( const int argc, const char * const argv[] )
           if( program_mode == m_compress )
             set_c_outname( input_filename, volume_size > 0 );
           else set_d_outname( input_filename, eindex );
-          outfd_mode = usr_rw;
-          if( !open_outstream( force ) )
+          if( !open_outstream( force, false ) )
             {
             if( retval < 1 ) retval = 1;
             close( infd ); infd = -1;
@@ -1015,17 +1045,20 @@ int main( const int argc, const char * const argv[] )
         }
       }
 
-    if( !check_tty( infd, program_mode ) ) return 1;
-
-    if( output_filename[0] && !to_stdout && program_mode != m_test )
-      delete_output_on_interrupt = true;
-    in_statsp = input_filename[0] ? &in_stats : 0;
     Pp_set_name( &pp, input_filename );
+    if( !check_tty( pp.name, infd, program_mode ) )
+      {
+      if( retval < 1 ) retval = 1;
+      if( program_mode == m_test ) { close( infd ); infd = -1; continue; }
+      cleanup_and_fail( retval );
+      }
+
+    in_statsp = input_filename[0] ? &in_stats : 0;
     if( program_mode == m_compress )
-      tmp = compress( member_size, volume_size, &encoder_options, infd,
-                      &pp, in_statsp );
+      tmp = compress( member_size, volume_size, infd, &encoder_options, &pp,
+                      in_statsp );
     else
-      tmp = decompress( infd, &pp, program_mode == m_test );
+      tmp = decompress( infd, &pp, ignore_trailing, program_mode == m_test );
     if( tmp > retval ) retval = tmp;
     if( tmp && program_mode != m_test ) cleanup_and_fail( retval );
 
